%{
uses lexlib, yacclib;

type
    ministring= string [7];
    tipotipo= record
                valor: ministring;
                tipo: integer;
              end;
    tiporegistro= record
                    valor: ministring;
                    case clase: integer of
                         2,3: (tipo: integer);
                         4: (parametro: array [1..5] of integer);
                  end;
    (******   clase:                       tipo:              
           1: constante                tipo simple:      1: integer
           2: tipo                     constructores:    1: array
           3: variable                                   2: puntero
           4: procedimiento
     ******************)

    tipotabla= array [1..100] of tiporegistro;

    tipolista= array [1..20] of integer; 

var
   tabla: tipotabla;  (* Tabla con los identificadores *)
   lista: tipolista;  (* Lista con la posicin de los bloques en la tabla *)
   lin, num, cont: integer;
   actualtabla: integer; (* Posicin de la tabla  *)
   actuallista: integer; (* Posicin de la lista  *)

(************************************************************************)
(*    Introduce un identificador en la tabla                            *)
(************************************************************************)
procedure InsertarTabla (reg: tiporegistro);
begin
     actualtabla:=actualtabla+1;
     tabla[actualtabla]:= reg;
end;

(************************************************************************)
(*    Imprime los valores de la tabla en pantalla                       *)
(************************************************************************)
procedure ImprimirTabla;
var
   i,j : integer;
   aux: string;
begin

     writeln (' *********    TABLA     ***********');
     for i:= 1 to actualtabla do
     begin
          if Tabla[i].clase=4 then
          begin
             write (Tabla[i].valor,'  ',Tabla[i].clase,'  ');
             for j:=1 to 4 do
                 write (Tabla[i].parametro[j]);
             writeln (Tabla[i].parametro[j]);
          end
          else
             writeln (Tabla[i].valor,'  ',Tabla[i].clase,'  ',Tabla[i].tipo);
     end;
end;

(************************************************************************)
(*    Busca un identificador en la tabla y devuelve su tipo             *)
(************************************************************************)
function DevolverTipo (reg: ministring) : integer;
var
   resul, i: integer;
   encontrado: boolean;
begin
     if reg='INTEGER' then
     begin
          resul:=1;
     end
     else
     begin
        i:=actualtabla+1;
        encontrado:=false;
        repeat
              i:=i-1;
              if Tabla[i].valor= reg then
              begin
                   encontrado:=true;
              end;
        until (encontrado) or (i=1);
        if (encontrado) and ((tabla[i].clase=2) or (tabla[i].clase=3) or (tabla[i].clase=1)) then
           resul:= (tabla[i].tipo)
        else
        begin
          if not encontrado then
             writeln ('ERROR: Variable ',reg,' no declarada')
          else
             writeln ('ERROR: Tipo de ',reg,' no vlido');
          halt;
        end;
     end;
     DevolverTipo:= resul;
end;

(************************************************************************)
(*  Busca un identificador en la tabla y devuelve su clase              *)
(************************************************************************)
function DevolverClase (reg: ministring) : integer;
var
   resul, i: integer;
   encontrado: boolean;
begin
     if reg='INTEGER' then
        Resul:= 2
     else
     begin
        i:=actualtabla+1;
        encontrado:=false;
        repeat
           i:=i-1;
           if Tabla[i].valor= reg then
           begin
                encontrado:=true;
           end;
        until (encontrado) or (i=1);
        if (encontrado) and ((tabla[i].clase=2) or (tabla[i].clase=3) or (tabla[i].clase=1)) then
            resul:= (tabla[i].clase)
        else
        begin
            if not encontrado then
                writeln ('ERROR: Variable ',reg,' no declarada')
            else
                writeln ('ERROR: Tipo de ',reg,' no vlido');
                halt;
        end;
     end;
     DevolverClase:= resul;
end;


(************************************************************************)
(*   Pone el tipo de variables ya introducidas en la tabla              *)
(************************************************************************)
procedure PonerTipo (t: integer);
var
   i: integer;
begin
     cont:=cont+1;
     for i:=cont to actualTabla do
        Tabla[i].tipo:= t;
     cont:=actualtabla;
end;

(************************************************************************)
(*   Introduce el tipo de los parametros de un procedimiento            *)
(************************************************************************)
procedure InsertarParam (num: integer; t: integer);
var
   i: integer;
begin
   for i:=1 to num do
   begin
        Tabla[lin].parametro[cont]:=t;
        Tabla[lin+cont].tipo:= t;
        cont:=cont+1;
   end;
end;

(************************************************************************)
(*  Inserta en la tabla los parametros del procedimiento como variables *)
(************************************************************************)
procedure InsertarParTabla (reg: tiporegistro);
begin
     actualtabla:= actualtabla+1;
     Tabla[actualtabla].valor:= reg.valor;
     Tabla[actualtabla].clase:= 3;
end;

(************************************************************************)
(*  Borra un bloque de cdigo de la tabla                               *)
(************************************************************************)
procedure BorrarVariables;
var
   i: integer;
begin
     for i:=lista[actuallista] to actualtabla do
          Tabla[i].valor:='';
     actualtabla:= lista[actuallista]-1;
     lista[actuallista]:=0;
     actuallista:=actuallista-1;
end;

(************************************************************************)
(*Busca el nombre de un procedimiento en la tabla y devuelve su posicin*)
(************************************************************************)
function BuscarProcedimiento(nombre: ministring): integer;
var
   i: integer;
   encontrado: boolean;

begin
     i:= actualtabla+1;
     encontrado:=false;
     repeat
           i:=i-1;
           if (Tabla[i].valor= nombre) and (Tabla[i].clase=4) then
                encontrado:= true;
     until (encontrado) or (i=0);
     if encontrado then
        BuscarProcedimiento:= i
     else
     begin
          writeln ('ERROR: Nombre de procedimiento incorrecto');
          halt;
     end;
end;

(************************************************************************)
(* Comprueba el tipo de los parametros en las llamadas a procedimientos *)
(************************************************************************)
procedure ComprobarTipo (cla, tip: integer);
begin
     if (cla<>1) and (cla<>3) then
     begin
        writeln ('ERROR: Argumento no vlido');
        halt;
     end;
     if tip<>Tabla[num].Parametro[cont] then
     begin
        writeln ('ERROR: Tipo de argumento incorrecto');
        halt;
     end;
     cont:= cont+1;
end;


%}

%token _PROGRAM _CONST _TYPE _VAR _ARRAY _OF
%token _PROCEDURE _BEGIN _END
%token PUNTOPUNTO ASIGNACION
%token <tiporegistro> IDENT
%token <ministring> ENTERO

%type <tipotipo> tipo
%type <tiporegistro> referencia expresion

%%

programa : _PROGRAM IDENT '(' listaIdentificadores ')' ';'
                bloqueConstantes bloqueTipos bloqueVariables
                bloqueProcedimientos sentenciaSecuencial '.' {ImprimirTabla}
          ; 

listaIdentificadores : listaIdentificadores ',' IDENT {$3.clase:=3; InsertarTabla ($3);} 
                     | IDENT {$1.clase:=3; InsertarTabla ($1);}
                     ;

bloqueConstantes : _CONST listaConstantes 
                 |
                 ;

listaConstantes: listaConstantes IDENT '=' ENTERO ';' { $2.clase:=1; $2.tipo:=1; InsertarTabla ($2)}
               | IDENT '=' ENTERO ';' { $1.clase:=1; $1.tipo:=1; InsertarTabla ($1)} 
               ;

bloqueTipos : _TYPE listaTipos
            |
            ;

listaTipos : listaTipos IDENT '=' tipo ';' {$2.clase:=2;
                                            $2.tipo:=$4.tipo; InsertarTabla ($2)}
           | IDENT'=' tipo ';' {$1.clase:=2;
                                $1.tipo:=$3.tipo; InsertarTabla ($1)}
           ;

tipo : IDENT {$$.tipo:= DevolverTipo($1.valor);
              if DevolverClase ($1.valor)<>2 then begin
              writeln ('ERROR: Debe haber un tipo'); halt; end;}
     | _ARRAY '[' ENTERO PUNTOPUNTO ENTERO ']' _OF tipo
                {$$.tipo:= 1+(3*DevolverTipo($8.valor));}
     | '^' IDENT {$$.tipo:= 2+(3*DevolverTipo($2.valor));
                  if DevolverClase ($2.valor)<>2 then begin
                  writeln ('ERROR: Debe haber un tipo'); halt; end;}
     ;

bloqueVariables : _VAR {cont:=actualtabla;} listaVariables
                |
                ;

listaVariables : listaVariables listaIdentificadores ':' tipo ';' {PonerTipo ($4.tipo);}
               | listaIdentificadores ':' tipo ';'{PonerTipo ($3.tipo);}
               ;

bloqueProcedimientos : listaProcedimientos
                     |
                     ;

listaProcedimientos : listaProcedimientos procedimiento
                    | procedimiento
                    ;

procedimiento : _PROCEDURE IDENT { $2.clase:=4; InsertarTabla ($2); lin:= actualtabla;
                actuallista:=actuallista+1; lista[actuallista]:=actualtabla+1;}
                parametros ';'
                bloqueConstantes bloqueTipos
                bloqueVariables bloqueProcedimientos
                sentenciaSecuencial ';'  { BorrarVariables;}
              ;

parametros : {cont:=1} '(' listaParametros ')'
           ;

listaParametros : listaParametros ';' parametro
                | parametro 
                ;

parametro : _VAR {num:=1} listaIdentificadoresPar ':' tipo {InsertarParam (num, $5.tipo);}
          | {num:=1} listaIdentificadoresPar ':' tipo {InsertarParam (num, $4.tipo);}
          ;

listaIdentificadoresPar: listaIdentificadoresPar ',' IDENT {num:=num+1;InsertarParTabla($3);}
                        | IDENT {InsertarParTabla ($1);}
                        ;

sentenciaSecuencial : _BEGIN listaSentencias _END
                    ;

listaSentencias : listaSentencias ';' sentencia
                | sentencia
                ;

sentencia : referencia ASIGNACION expresion {If $1.tipo<>$3.tipo then
                                       begin write ('ERROR: Asignacin incorrecta');halt; end;
                                       if $1.clase<>3 then begin write ('ERROR: Debe asignarse a una variable');
                                       halt; end;}
          | IDENT { write ('ERROR: Sentencia Incorrecta'); halt;}
          | IDENT {cont:=1; num:= BuscarProcedimiento($1.valor);}
                '(' listaArgumentos ')' 
          |
          ;

referencia: IDENT {$$.tipo:= DevolverTipo ($1.valor);
                   $$.clase:= DevolverClase ($1.valor);}
          | referencia '[' expresion ']' {$$.tipo:= $1.tipo div 3;
                                          If $3.tipo<>1 then begin
                                          write ('ERROR: Indexacin de array incorrecta'); halt; end;
                                          $$.clase:= $1.clase;}
          | referencia '^' {$$.tipo:= $1.tipo div 3;
                            $$.clase:= $1.clase;}
          ;

expresion : referencia {$$.tipo:= $1.tipo;
                        $$.clase:= $1.clase}
          | ENTERO {$$.tipo:=1; $$.clase:=1;}
          ;

listaArgumentos : listaArgumentos ',' referencia {ComprobarTipo ($3.clase, $3.tipo);}
                | referencia {ComprobarTipo ($1.clase, $1.tipo);}
                ;

%%

{$I praclex.pas}

begin
     actualtabla:=0;
     lista[1]:=1;
     actuallista:=1;

     if yyparse=0 then
        writeln ('CORRECTO');
end.
